//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#include "stdafx.h"
#include "BaseWrap.h"
#include "PropertyNameIndex.h"
#include "spatial/SltGeomUtils.h"

#include <float.h>


namespace fdo
{

//=============================================================================
//=============================================================================
// ColumnDef
//=============================================================================
//=============================================================================

ColumnDef::~ColumnDef()
{
    delete _csys;
}


void ColumnDef::SetCoordSys(const CoordSys* csys)
{
    delete _csys;
    if (csys)
        _csys = new CoordSys(*csys);
    else
        _csys = NULL;
}

void ColumnDef::Initialize(const wchar_t* name, FdoDataType dataType, const CoordSys* csys)
{
    _name = name;
    _mbname = W2A_SLOW(name);
    _dataType = dataType;
    SetCoordSys(csys);
}


//=============================================================================
//=============================================================================
// TableDef
//=============================================================================
//=============================================================================

TableDef::TableDef()
{
    _nameToIndex = NULL;
    _cols = NULL;
    _ncols = 0;
    _idindex = -1;
    _geomindex = -1;
}


void TableDef::Initialize(const wchar_t* name, ColumnDef* defs, int ndefs, int idindex, int geomindex)
{
    _name = name;

    //we take ownership
    delete [] _cols;
    _cols = defs;
    _ncols = ndefs;
    _idindex = idindex;
    _geomindex = geomindex;
    
    delete _nameToIndex;
    _nameToIndex = new PropertyNameIndex();

    for (int i=0; i<ndefs; i++)
    {
        _nameToIndex->Add(_cols[i].Name(), i);
    }

    //build the name-index map
    _nameToIndex->Prepare();
}

TableDef::~TableDef()
{
    delete _nameToIndex;
    delete [] _cols;
}


int TableDef::IndexOf(const wchar_t* name) const
{
    int ret = _nameToIndex->GetIndex(name);
    if (ret == -1)
        throw; //this should not happen -- looking for a property that doesn't exist
    return ret;
}



//=============================================================================
//=============================================================================
// CoordSysDef
//=============================================================================
//=============================================================================

//ownership of input CoordSys list not transferred
CoordSysDef::CoordSysDef(const CoordSys* cslist, int ncs)
{
    _cslist = new CoordSys*[ncs];
    _ncs = ncs;
    _nameToIndex = new PropertyNameIndex();

    for (int i=0; i<ncs; i++)
    {
        _cslist[i] = new CoordSys(cslist[i]);
        _nameToIndex->Add(cslist[i].Name(), i);
    }

    _nameToIndex->Prepare();
}

CoordSysDef::~CoordSysDef()
{
    delete _nameToIndex;

    for (int i=0; i<_ncs; i++) 
        delete _cslist[i];
    delete [] _cslist;
}

int CoordSysDef::IndexOf(const wchar_t* name) const
{
    return _nameToIndex->GetIndex(name);
}



//=============================================================================
//=============================================================================
// DatabaseDef
//=============================================================================
//=============================================================================

DatabaseDef::DatabaseDef(const wchar_t* name, TableDef* tdef, int ntbl, CoordSysDef* csdef)
{
    _name = name ? name : L"";
    _rowdefs = tdef;
    _count = ntbl;
    _csdef = csdef;

    _nameToIndex = new PropertyNameIndex();

    for (int i=0; i<ntbl; i++)
    {
        _nameToIndex->Add(_rowdefs[i].Name(), i);
    }

    _nameToIndex->Prepare();
}

DatabaseDef::~DatabaseDef()
{
    delete _csdef;
    delete _nameToIndex;
    delete [] _rowdefs;
}

int DatabaseDef::IndexOf(const wchar_t* name) const
{
    return _nameToIndex->GetIndex(name);
}

//=============================================================================
//=============================================================================
// Row
//=============================================================================
//=============================================================================

Row::Row(const TableDef* ri)
{
    _info = ri;
    _vals = new Value[ri->ColumnCount()];
}

Row::~Row()
{
    delete [] _vals;
}

void Row::GeomBBOX(int i, double bbox[4]) const
{
    if (_vals[i].Type() == ColumnDef::DataType_Geom)
    {
        GetFgfExtents((const unsigned char*)_vals[i].AsBlob(), _vals[i].BlobLen(), bbox);
    }
    else
    {
        bbox[0] = bbox[1] = DBL_MAX;
        bbox[2] = bbox[3] = -DBL_MAX;
    }
}

//=============================================================================
//=============================================================================
// Database
//=============================================================================
//=============================================================================


void Database::ClearSchema()
{
    if (_schema)
    {
        for (int i=0; i<_schema->Count(); i++)
            delete _tables[i];
        delete [] _tables;
        _tables = NULL;
        delete _schema;
        _schema = NULL;
    }
}


} 