//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//


#pragma once


#include "FdoWrapDll.h"
#include "Value.h"
#include <assert.h>

class PropertyNameIndex;

namespace fdo
{


enum
{
    INVALID_FID = 0
};


//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    CoordSys
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//Simple container for FDO-style coord sys definitions
class FDOWRAP_API CoordSys
{
public:
    CoordSys(const wchar_t* name, const wchar_t* code, const wchar_t* wkt)
    {
        if (name) _name = name;
        if (code) _code = code;
        if (wkt) _wkt = wkt;
    }

    const wchar_t* Name() const { return _name.c_str(); }
    const wchar_t* Code() const { return _code.c_str(); }
    const wchar_t* Wkt() const { return _wkt.c_str(); }

private:
    std::wstring _name;
    std::wstring _code;
    std::wstring _wkt;
};

//Simple collection of CoordSys objects, providing access by name and index
class FDOWRAP_API CoordSysDef
{
public:
    CoordSysDef(const CoordSys* cslist, int ncs);
    ~CoordSysDef();

    int Count() const { return _ncs; }
    const CoordSys& operator[] (int i) const { return *_cslist[i]; }
    const CoordSys* operator[] (const wchar_t* name) const 
    { 
        int ret = IndexOf(name);
        if (ret == -1)
            return NULL;
        return _cslist[ret]; 
    }
    int IndexOf(const wchar_t* name) const;
    
private:
    CoordSys** _cslist;
    int _ncs;
    PropertyNameIndex* _nameToIndex; 
};


//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    ColumnDef -- information about a single column (equivalent to an FDO property definition)
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class FDOWRAP_API ColumnDef
{
public:
    enum GeomDataTypeHolder
    {
        Fdo_Boolean = 0,
        Fdo_Byte,
        Fdo_DateTime,
        Fdo_Decimal,
        Fdo_Double,
        Fdo_Int16,
        Fdo_Int32,
        Fdo_Int64,
        Fdo_Single,
        Fdo_String,
        Fdo_BLOB,
        Fdo_CLOB,

        DataType_Geom = -1,
        DataType_Unknown = -2
    };

    ColumnDef()
    {
        _dataType = (FdoDataType)INT_MIN;
        _csys = NULL;
    }
    ~ColumnDef();

    void Initialize(const wchar_t* name, FdoDataType dataType, const CoordSys* csys = NULL);

    const wchar_t* Name() const { return _name.c_str(); }
    const char* NameUTF8() const { return _mbname.c_str(); }
    FdoDataType DataType() const { return _dataType; } //returns values from the FdoDataType enumeration
    const CoordSys* CSys() const { return  _csys; } //NULL if not geometry, may be not NULL if geometry

    //Need this setter since this is initialized as a post-process step... Sadly.
    void SetCoordSys(const CoordSys* csys);

private:
    //Ball out of play here
    ColumnDef(const ColumnDef&);
    ColumnDef& operator= (const ColumnDef&);

    std::wstring _name;
    std::string _mbname;
    FdoDataType _dataType;
    CoordSys* _csys;
};

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    TableDef - defines the columns of a table, same as FdoClassDefinition
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class CoordSysDef;

class FDOWRAP_API TableDef
{
public:
    TableDef();
    ~TableDef();

    void Initialize(const wchar_t* name, ColumnDef* defs, int ndefs, int idindex = -1, int geomindex = -1);

    int IndexOf(const wchar_t*) const;

    const ColumnDef& ID() const { return _cols[IndexOfID()]; }
    int IndexOfID() const { return _idindex; }

    const ColumnDef& Geom() const { return _cols[IndexOfGeom()]; }
    int IndexOfGeom() const { return _geomindex; }

    const ColumnDef& operator[] (int i) const { return _cols[i]; }
    int ColumnCount() const { return _ncols; }

    const wchar_t* Name() const { return _name.c_str(); }

private:

    PropertyNameIndex* _nameToIndex; 
    ColumnDef* _cols;
    int _ncols;
    int _idindex;
    int _geomindex;
    std::wstring _name;

private:
    //Verboten
    TableDef(const TableDef&);
    TableDef& operator= (const TableDef&);
};



//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    Row - A data row -- contains attribute values for a single feature
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class FDOWRAP_API Row
{
public:

    Row(const TableDef*);
    ~Row();

    //Column access by index
    Value& operator[] (int i) { return _vals[i]; }
    const Value& operator[] (int i) const { return _vals[i]; }

    //Column access by name
    Value& operator[] (const wchar_t* name)
    {
        int ret = _info->IndexOf(name);

        assert(ret != -1);

        return _vals[ret];
    }
    const Value& operator[] (const wchar_t* name) const
    {
        int ret = _info->IndexOf(name);

        assert(ret != -1);

        return _vals[ret];
    }

    //The ID of the row, or IVALID_FID if no ID
    __int64 ID() const
    {
        int i = _info->IndexOfID();
        if (i == -1) //detects the case of no ID property -- ideally this would not be necessary
            return INVALID_FID; 
        return _vals[i].AsInt();
    }

    //Shortcut to the geometry value
    const void* Geom() const
    {
        int i = _info->IndexOfGeom();
        if (i == -1) //detects the case of no geometry
            return NULL;
        return _vals[i].AsBlob();
    }

    //Shortcut to length of geometry array
    int GeomLen() const 
    {
        int i = _info->IndexOfGeom();
        if (i == -1) //detects the case of no geometry
            return NULL;
        return _vals[i].BlobLen();
    }

    //Computes and returns the bounding box of a geometry column
    //Returns invalid bounds on failure (invalid defined as min > max)
    void GeomBBOX(int i, double bbox[4]) const;

    //The definition of the row
    const TableDef& Def() const 
    { 
        return *_info; 
    }

private:
    Value* _vals;
    const TableDef* _info;

private:
    //Here be dragons.
    Row(const Row&);
    Row& operator= (const Row&);

};



//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    DatabaseDef -- represents an FDO schema in TableDef format
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class FDOWRAP_API DatabaseDef
{
public:

    DatabaseDef(const wchar_t* name, TableDef* tdef, int ntbl, CoordSysDef* csdef);
    ~DatabaseDef();

    //Access to table defs by name and index
    int IndexOf(const wchar_t* name) const;
    const TableDef* operator[] (const wchar_t* name) const
    {
        int ret = IndexOf(name);
        if (ret == -1)
            return NULL; //should never happen
        return &_rowdefs[ret];
    }
    const TableDef& operator[] (int i) const { return _rowdefs[i]; }
    int Count() const { return _count; }

    const wchar_t* Name() const { return _name.c_str(); }

    //Coodinate system definitions
    const CoordSysDef& CSDef() const { return *_csdef; }

private:

    PropertyNameIndex* _nameToIndex; 
    TableDef* _rowdefs;
    int _count;
    std::wstring _name;
    CoordSysDef* _csdef;
};





//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//  Table -- (FDO class, OGR layer), implements query, insert, delete, update
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


class FDOWRAP_API Table
{
public:

    virtual ~Table() {}

    //Modification operations
    //TODO: BeginBulkInsert/EndBulkInsert
    virtual bool Insert(const Row& row, __int64* newfid = NULL, bool skipID = false) = 0;

    virtual bool Update() { return Update(At()); }
    virtual bool Update(const Row& row) = 0;

    virtual bool Delete(__int64 fid) = 0;

    //Query operations
    virtual void        Query       () = 0; //get them all
    virtual void        Query       (const double bbox[4], FdoFilter* attributeFilter = NULL) = 0; //simple bbox query cursor
    virtual const Row*  Query       (__int64 id) = 0; //query for specific feature
    virtual const Row*  Query       (int id) { return Query((__int64)id); }//query for specific feature
    virtual void        Query       (FdoFilter*) = 0; //custom filter
    virtual void        QueryRange  (__int64 minID, __int64 maxID) = 0;
    virtual void        QueryRandomAccess() = 0; //random access cursor

    virtual void EndQuery() = 0;

    //Aggregate operations
    virtual void GetExtent(double bbox[4]) = 0;
    virtual __int64 GetRowCount() = 0;
    
    //reading of query results
    virtual const Row* Next() = 0; //advance cursor, returns current row, or NULL if done
    virtual const Row& At() = 0; //row at current cursor position
    //virtual const Row& At(__int64 i) = 0; //random access

    //TODO: make this more friendly
    //Row& WR() { return *_row; } //writable row for use by inserts, updates

    const TableDef& Def() const { return *_ri; }

protected:
    const TableDef* _ri;
};



//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//  A whole database, containing possibly many tables. 
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


class FDOWRAP_API Database
{
public:
    Database() 
        : _tables(NULL), _schema(NULL)
    {}

    virtual ~Database() {}

    virtual bool Connect(const wchar_t* connStr, bool readOnly) = 0;
    
    virtual void Disconnect() = 0;

    virtual bool SetSchema(const DatabaseDef& def) = 0;


    //Access to tables by name and index
    Table& operator[] (int i) { return *_tables[i]; }
    Table* operator[] (const wchar_t* name) 
    { 
        int r = Def().IndexOf(name);
        if (r == -1)
            return NULL; //should not happen
        return _tables[r];
    }
    int Count() const { return _schema ? _schema->Count() : 0; }

    //The database schema definition
    const DatabaseDef& Def() const { return *_schema; }
    const wchar_t* ConnStr() const { return _connStr.c_str(); }

protected:

    void ClearSchema();

    Table** _tables;
    DatabaseDef* _schema;
    std::wstring _connStr;
};



}

