//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#include "stdafx.h"
#include "DbPool.h"
#include "BaseWrap.h"
#include "FdoWrap.h"
#include "OgrWrap.h"
#include "mutex.h"
#include <time.h>

namespace fdo
{

//connection pool record -- contains database pointer
//plus extra timing information/status
struct PoolRec
{
    PoolRec(Database* db, time_t tm)
    {
        _db = db;
        _stamp = tm;
    }

    Database* _db;
    time_t _stamp;
};

typedef std::map<std::wstring, std::vector<PoolRec> > ConnPool;


//makes sure connections are closed on exit
struct PoolHolder
{
    ~PoolHolder()
    {
        DbPool::CloseAllFreeConnections();
    }

    ConnPool freePool;
    ConnPool busyPool;
    Mutex    mutex;
};

//The One instance of the connection pool
PoolHolder g_pool;



Database* DbPool::GetConnection(const wchar_t* connStr)
{
    ScopedLock scc(g_pool.mutex);

    ConnPool::iterator it = g_pool.freePool.find(connStr);

    //no connection for this string in the pool -- make one
    if (it == g_pool.freePool.end() || it->second.size() == 0)
    {
#if !EXCLUDE_FDO
        //Try to open the connection with FDO
        Database* db = new DatabaseFdo();
        if (db->Connect(connStr, false)) //TODO: proper passing of ReadOnly
        {
            return db;
        }
        else
        {
            delete db;
        }
#endif

#if !EXCLUDE_OGR
        //FDO failed, try OGR
        db = new DatabaseOgr();
        if (db->Connect(connStr, true)) //TODO: proper passing of ReadOnly
        {
            return db;
        }
        else
        {
            delete db;
        }
#endif

        return NULL;
    }
    else
    {
        PoolRec rec = it->second.back();
        it->second.pop_back();
        return rec._db;
    }

    return NULL;
}


void DbPool::ReleaseConnection(Database* db)
{
    ScopedLock scc(g_pool.mutex);

    std::vector<PoolRec>& vec = g_pool.freePool[db->ConnStr()];

    vec.push_back(PoolRec(db, time(NULL)));

}


void DbPool::CloseAllFreeConnections()
{
    ScopedLock scc(g_pool.mutex);
    
    for (ConnPool::iterator it = g_pool.freePool.begin(); it != g_pool.freePool.end(); ++it)
    {
        while (it->second.size())
        {
            delete it->second.back()._db;
            it->second.pop_back();
        }
    }
}






DbHandle::DbHandle(Database* db)
{
    _db = db;
}

DbHandle::~DbHandle()
{
    if (_db)
        DbPool::ReleaseConnection(_db);
}

DbHandle& DbHandle::operator=(Database* db)
{
    if (_db)
        DbPool::ReleaseConnection(_db);

    _db = db;

    return *this;
}


}