//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#include "stdafx.h"

#if !EXCLUDE_FDO

#include "FdoWrap.h"

#include "PropertyNameIndex.h"
#include "spatial/SltGeomUtils.h"

namespace fdo
{

//=============================================================================
// Ugly utility code
//=============================================================================

typedef FdoIConnection* (*createFunc)();

FdoIConnection* CreateConnection(const wchar_t* provider)
{
#ifdef _WIN32
    HMODULE h = LoadLibrary(provider);
    if (!h) goto fail_load;

    FARPROC f = GetProcAddress(h, "CreateConnection");
    if (!f) goto fail_entry;

    return ((createFunc)f)();
#else
    char mbsprov[PATH_MAX];
    wcstombs(mbsprov, provider);
    
    void* h = dlopen(mbsprov, RTLD_LAZY);
    if (!h) goto fail_load;
    
    void* f = dlsym(h, "CreateConnection");
    if (!f) goto fail_entry;    

    return ((createFunc)f)();
#endif

fail_load:
    fprintf(stderr, "Failed to load provider: %ls\n", provider);
    return NULL;

fail_entry:
    fprintf(stderr, "Failed to find FDO entry point: %ls\n", provider);
    return NULL;
}

FdoIConnection* GetFdoCon(const wchar_t* srcfile, bool open)
{
    std::wstring connstr;

    _ASSERT(wcslen(srcfile) > 4);

    std::wstring fname = srcfile;
    size_t index = fname.find_last_of(L'.');

    if (index == std::wstring::npos)
    {
        return NULL;
    }

    std::wstring ext = fname.substr(index);
    std::wstring provider;

    if (_wcsicmp(ext.c_str(), L".sdf") == 0)
    {
        provider = L"SdfProvider"; 
        connstr = std::wstring(L"File=") + srcfile + std::wstring(L";ReadOnly=false;");
    }
    /*
    else if (_wcsicmp(ext.c_str(), L".shp") == 0)
    {
        provider = L"ShpProvider";
        connstr = std::wstring(L"DefaultFileLocation=") + srcfile + std::wstring(L";");
    }
    */
    else if (_wcsicmp(ext.c_str(), L".sdx") == 0 || wcscmp(ext.c_str(), L".db") == 0 || wcscmp(ext.c_str(), L".sqlite") == 0 || wcscmp(ext.c_str(), L".alxdb") == 0)
    {
        provider = L"SQLiteProvider";
        connstr = std::wstring(L"File=") + srcfile + std::wstring(L";UseFdoMetadata=true;");
    }
    else
    {
        return NULL;
    }

    FdoIConnection* conn = CreateConnection(provider.c_str());

    if (!conn)
        return NULL;

    //open an FDO connection to the file
    conn->SetConnectionString(connstr.c_str());

    try
    {
        if (open)
            conn->Open();
    }
    catch (FdoException* e)
    {
        ALX_ERROR(e->GetExceptionMessage());
        e->Release();
    }

    return conn;
}



//=============================================================================
//=============================================================================
// TableFdo
//=============================================================================
//=============================================================================


static FdoClassDefinition* TableDefToFdoClass(const TableDef& td) 
{
    FdoClassDefinition* fc;

    if (td.IndexOfGeom() != -1)
    {
        fc = FdoFeatureClass::Create(td.Name(), NULL);
    }
    else
    {
        fc = FdoClass::Create(td.Name(), NULL);
    }

    int igeom = td.IndexOfGeom();
    int iid = td.IndexOfID();

    FdoPtr<FdoPropertyDefinitionCollection> pdc = fc->GetProperties();

    for (int i=0; i<td.ColumnCount(); i++)
    {
        const ColumnDef& cd = td[i];
        
        if (cd.DataType() == ColumnDef::DataType_Geom)
        {
            FdoPtr<FdoGeometricPropertyDefinition> gpd = FdoGeometricPropertyDefinition::Create(cd.Name(), NULL);
            gpd->SetSpatialContextAssociation(cd.CSys() ? cd.CSys()->Name() : NULL);
            pdc->Add(gpd);

            if (i == igeom)
            {
                ((FdoFeatureClass*)fc)->SetGeometryProperty(gpd);
            }
        }
        else
        {
            FdoPtr<FdoDataPropertyDefinition> dpd = FdoDataPropertyDefinition::Create(cd.Name(), NULL);
            dpd->SetDataType(cd.DataType());
            pdc->Add(dpd);

            if (i == iid)
            {
                dpd->SetIsAutoGenerated(true); //TODO: need to remember this in the ColumnDef
                FdoPtr<FdoDataPropertyDefinitionCollection> idpdc = fc->GetIdentityProperties();
                idpdc->Add(dpd);
            }
        }
    }

    return fc;
}

static void FdoClassToTableDef(TableDef& td, FdoClassDefinition* fc, const CoordSysDef& csdef)
{
    FdoPtr<FdoPropertyDefinitionCollection> pdc = fc->GetProperties();
    FdoPtr<FdoDataPropertyDefinitionCollection> idpdc = fc->GetIdentityProperties();

    int idindex = -1;
    int geomindex = -1;
    std::wstring idname;
    std::wstring geomname;

    //id property name
    for (int i=0; i<idpdc->GetCount(); i++)
    {
        FdoPtr<FdoDataPropertyDefinition> idpd = idpdc->GetItem(i);
        idname = idpd->GetName();
        //should be just one identity property
        break;
    }

    //geom property name
    if (fc->GetClassType() == FdoClassType_FeatureClass)
    {
        FdoPtr<FdoGeometricPropertyDefinition> gpd = ((FdoFeatureClass*)fc)->GetGeometryProperty();
        geomname = gpd->GetName();
    }

    //build the name-index map and also collect info about each column
    int nprops = pdc->GetCount();
    ColumnDef* cols = new ColumnDef[nprops];

    for (int i=0; i<nprops; i++)
    {
        FdoPtr<FdoPropertyDefinition> pd = pdc->GetItem(i);

        if (pd->GetPropertyType() == FdoPropertyType_DataProperty)
        {
            FdoDataPropertyDefinition* dpd = (FdoDataPropertyDefinition*)(pd.p);

            cols[i].Initialize(dpd->GetName(), dpd->GetDataType());

            if (idindex == -1 && wcscmp(dpd->GetName(), idname.c_str()) == 0)
                idindex = i;
        }
        else if (pd->GetPropertyType() == FdoPropertyType_GeometricProperty)
        {
            FdoGeometricPropertyDefinition* gpd = (FdoGeometricPropertyDefinition*)(pd.p);
            
            cols[i].Initialize(gpd->GetName(), (FdoDataType)ColumnDef::DataType_Geom, csdef[gpd->GetSpatialContextAssociation()]);

            if (geomindex == -1 && wcscmp(gpd->GetName(), geomname.c_str()) == 0)
                geomindex = i;
        }
        else
        {
            //just leave the ColumnDef with empty name
        }
    }

    td.Initialize(fc->GetName(), cols, nprops, idindex, geomindex);
}


//Sets up an insert command for a given feature class. The Insert command
//contains FdoPropertyValues for each property, including the id, which we
//will override into the destination. 
static void PopulateInsertCommand(FdoIInsert *insert, const TableDef& ri)
{
    FdoPtr<FdoPropertyValueCollection> pvc = insert->GetPropertyValues();

    insert->SetFeatureClassName(ri.Name());

    for (int i=0, iEnd = ri.ColumnCount(); i<iEnd; i++)
    {
        //fprintf(stderr, "prop name : %ls\n", pd->GetName());

        FdoPtr<FdoPropertyValue> pv = FdoPropertyValue::Create(ri[i].Name(), NULL);
        pvc->Add(pv);

        int dt = ri[i].DataType();

        switch (dt)
        {
        case FdoDataType_BLOB:
            {
                FdoPtr<FdoBLOBValue> bv = FdoBLOBValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Boolean:
            {
                FdoPtr<FdoBooleanValue> bv = FdoBooleanValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Byte:
            {
                FdoPtr<FdoByteValue> bv = FdoByteValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_CLOB:
            {
                FdoPtr<FdoCLOBValue> bv = FdoCLOBValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_DateTime:
            {
                FdoPtr<FdoDateTimeValue> bv = FdoDateTimeValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Decimal:
            {
                FdoPtr<FdoDecimalValue> bv = FdoDecimalValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Double:
            {
                FdoPtr<FdoDoubleValue> bv = FdoDoubleValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Int16:
            {
                FdoPtr<FdoInt16Value> bv = FdoInt16Value::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Int32:
            {
                FdoPtr<FdoInt32Value> bv = FdoInt32Value::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Int64:
            {
                FdoPtr<FdoInt64Value> bv = FdoInt64Value::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_Single:
            {
                FdoPtr<FdoSingleValue> bv = FdoSingleValue::Create();
                pv->SetValue(bv);
            }
            break;
        case FdoDataType_String:
            {
                FdoPtr<FdoStringValue> bv = FdoStringValue::Create();
                pv->SetValue(bv);
            }
            break;
        case ColumnDef::DataType_Geom: //our special value indicating geom property
            {
                FdoPtr<FdoGeometryValue> bv = FdoGeometryValue::Create();
                pv->SetValue(bv);
            }
            break;
        }
    }
}

//Transfers feature data from our Row/Value collection format to FDO property values
//for use by FDO insert command. The literal values of the property value collection are
//already pre-extracted and listed in the given vector (for faster access)
static void BindPropVals(FdoPropertyValueCollection* pvc, const Row& row, bool skipID)
{
    for (int i=0, iEnd = row.Def().ColumnCount(); i<iEnd; i++)
    {
#ifdef _DEBUG
        //It would be a major hard-to-detect bug if the RowInfo and the FDO property value
        //collection ordered their lists in different ways. It should not happen since they are
        //based on the exact same FdoClassDefinition, but let's check anyway.
        {
            FdoPtr<FdoPropertyValue> pv = pvc->GetItem(i);
            FdoPtr<FdoIdentifier> ident = pv->GetName();
            if (wcscmp(ident->GetName(), row.Def()[i].Name()) != 0)
            {
                ALX_ERROR("FDO property value collection, RowInfo property order mismatch!");
                throw;
            }
        }
#endif

        int dataType = row.Def()[i].DataType();

        if (!row[i].IsValid() || row[i].IsNull())
            continue; //do nothing since NULL is the default for the FDO literal values being passed in


        FdoPtr<FdoPropertyValue> pv = pvc->GetItem(i);
        FdoPtr<FdoValueExpression> ve = pv->GetValue();
        FdoLiteralValue* flv = (FdoLiteralValue*)(ve.p);

        if (skipID && i == row.Def().IndexOfID())
        {
            //in case we want to leave the ID as null (for example when we want
            //automatically generated one during insert):
            ((FdoDataValue*)flv)->SetNull();
            continue;
        }

        switch (dataType)
        {
        case FdoDataType_BLOB:
            {
                int len = 0;
                const void* blob = row[i].AsBlob();
                if (blob)
                {
                    FdoPtr<FdoByteArray> ba = FdoByteArray::Create((FdoByte*)blob, row[i].BlobLen());
                    ((FdoBLOBValue*)flv)->SetData(ba);                        
                }
            }
            break;
        case FdoDataType_Boolean:
            {
                ((FdoBooleanValue*)flv)->SetBoolean(row[i].AsBool());
            }
            break;
        case FdoDataType_Byte:
            {
                ((FdoByteValue*)flv)->SetByte((FdoByte)row[i].AsInt());
            }
            break;
        case FdoDataType_CLOB:
            {
                const void* blob = row[i].AsBlob();
                if (blob)
                {
                    FdoPtr<FdoByteArray> ba = FdoByteArray::Create((FdoByte*)blob, row[i].BlobLen());
                    ((FdoCLOBValue*)flv)->SetData(ba);                        
                }
            }
            break;
        case FdoDataType_DateTime:
            {
                FdoDateTime dt;
                if (DateFromString(dt, row[i].AsString()))
                {
                    ((FdoDateTimeValue*)flv)->SetDateTime(dt);
                }
                else
                {
                    ((FdoDateTimeValue*)flv)->SetNull();
                }
            }
            break;
        case FdoDataType_Decimal:
            {
                ((FdoDecimalValue*)flv)->SetDecimal(row[i].AsReal());
            }
            break;
        case FdoDataType_Double:
            {
                ((FdoDoubleValue*)flv)->SetDouble(row[i].AsReal());                
            } 
            break;
        case FdoDataType_Int16:
            {
                ((FdoInt16Value*)flv)->SetInt16(row[i].AsInt()); //data truncation, but should not happen anyway
            }
            break;
        case FdoDataType_Int32:
            {
                ((FdoInt32Value*)flv)->SetInt32(row[i].AsInt()); //data truncation, but should not happen anyway
            }
            break;
        case FdoDataType_Int64:
            {
                ((FdoInt64Value*)flv)->SetInt64(row[i].AsInt()); 
            }
            break;
        case FdoDataType_Single:
            {
                ((FdoSingleValue*)flv)->SetSingle(row[i].AsReal()); //data truncation, but should not happen anyway
            }
            break;
        case FdoDataType_String:
            {
                ((FdoStringValue*)flv)->SetString(row[i].AsString()); 
            }
            break;
        case -1:
            {
                const void* blob = row[i].AsBlob();
                if (blob)
                {
                    FdoPtr<FdoByteArray> ba = FdoByteArray::Create((FdoByte*)blob, row[i].BlobLen());
                    ((FdoGeometryValue*)flv)->SetGeometry(ba);
                }
            }
            break;
        }
    }
}



TableFdo::TableFdo(FdoIConnection* db, const TableDef* ri)
{
    _ri = ri;
    _db = FDO_SAFE_ADDREF(db);
    _row = new Row(ri);
}

TableFdo::~TableFdo()
{
    if (_rdr)
        _rdr->Close();

    delete _row;
}

bool TableFdo::Insert(const Row& row, __int64* newfid, bool skipID)
{
    if (!_insert)
    {
        //one time set up the insert command
        _insert = (FdoIInsert*)_db->CreateCommand(FdoCommandType_Insert);
        PopulateInsertCommand(_insert, *_ri);
        _pvc = _insert->GetPropertyValues();
    }

    bool result = true;
    BindPropVals(_pvc, row, skipID);

    try
    {
        FdoPtr<FdoIFeatureReader> rdr = _insert->Execute();

        //read and set the newly inserted FID, if caller requested it
        if (newfid)
        {
            if (rdr->ReadNext())
            {
                int t = Def().ID().DataType();
                switch (t)
                {
                case FdoDataType_Int32 : *newfid = rdr->GetInt32(Def().ID().Name()); break;
                case FdoDataType_Int64 : *newfid = rdr->GetInt64(Def().ID().Name()); break;
                default: *newfid = INVALID_FID;
                }
            }
            else
            {
                *newfid = INVALID_FID;
            }
        }
    }
    catch (FdoException* e)
    {
        ALX_ERROR(e->GetExceptionMessage()); 
        e->Release();
        result = false;
    }

    //zero out the FDO property values collection in preparation for the next feature
    for (int i=0, iEnd=_pvc->GetCount(); i<iEnd; i++)
    {
        FdoPtr<FdoPropertyValue> pv = _pvc->GetItem(i);
        FdoPtr<FdoLiteralValue> flv = (FdoLiteralValue*)pv->GetValue();
        if (row.Def()[i].DataType() != ColumnDef::DataType_Geom)
            ((FdoDataValue*)flv.p)->SetNull();
        else
            ((FdoGeometryValue*)flv.p)->SetNullValue();
    }

    return result;
}

bool TableFdo::Delete(__int64 fid)
{
    if (!_delete)
    {
        //one time set up of the delete command
        _delete = (FdoIDelete*)_db->CreateCommand(FdoCommandType_Delete);
        _delete->SetFeatureClassName(_ri->Name());

        std::wstring idname = (*_ri)[_ri->IndexOfID()].Name();

        FdoPtr<FdoIdentifier> name = FdoIdentifier::Create(idname.c_str());
        _delid = FdoInt64Value::Create(-1); //this will be modified for every feature
        FdoPtr<FdoComparisonCondition> filter = FdoComparisonCondition::Create(name, FdoComparisonOperations_EqualTo, _delid);
        _delete->SetFilter(filter);
    }


    bool result = true;
    _delid->SetInt64(fid);

    try
    {
        int res = _delete->Execute();

        if (!res)
            fprintf(stderr, "Feature delete failed...\n");
    }
    catch (FdoException* e)
    {
        ALX_ERROR(e->GetExceptionMessage()); 
        e->Release();
        result = false;
    }

    return result;
}

bool TableFdo::Update(const Row& row)
{
    //For update, we also want insert initialized, since
    //we reuse the property value collection from insert

    if (!_insert)
    {
        //one time set up the insert command
        _insert = (FdoIInsert*)_db->CreateCommand(FdoCommandType_Insert);
        PopulateInsertCommand(_insert, *_ri);
        _pvc = _insert->GetPropertyValues();
    }

    if (!_update)
    {
        _update = (FdoIUpdate*)_db->CreateCommand(FdoCommandType_Update);
        _update->SetFeatureClassName(_ri->Name());

        if (!_idfilter)
        {
            FdoPtr<FdoIdentifier> name = FdoIdentifier::Create(_ri->ID().Name());
            _updid = FdoInt64Value::Create(-1); //this will be modified for every feature
            _idfilter = FdoComparisonCondition::Create(name, FdoComparisonOperations_EqualTo, _updid);
            _update->SetFilter(_idfilter);
        }
    }

    bool result;
    _updid->SetInt64(row.ID());

    //populate property values for update -- we will
    //reuse the propval objects created for insert
    BindPropVals(_pvc, row, true);

    //transfer from the insert collection to the update collection,
    //but only the properties which are defined in the input row
    FdoPtr<FdoPropertyValueCollection> upvc = _update->GetPropertyValues();
    for (int i=0; i<row.Def().ColumnCount(); i++)
    {
        if (i != row.Def().IndexOfID() && row[i].IsValid())
        {
            FdoPtr<FdoPropertyValue> pv = _pvc->GetItem(i);
            upvc->Add(pv);
        }
    }

    try
    {
        int count = _update->Execute();
        result = !!count;
    }  
    catch (FdoException* e)
    {
        ALX_ERROR(e->GetExceptionMessage()); 
        e->Release();
        result = false;
    }

    upvc->Clear(); //ready for next time

    return result;
}


static void CheckClassDef(FdoClassDefinition* fc, const TableDef& ri)
{
    //sanity check for the class definition of the returned
    //reader matching the one we think it is. For simple query
    //we have to make sure the count and names of properties match exactly.
    FdoPtr<FdoPropertyDefinitionCollection> pvc = fc->GetProperties();
    if (pvc->GetCount() != ri.ColumnCount())
        ALX_ERROR("FDO class definition does not match TableDef. Column values will mismatch!");
    for (int i=0, iEnd = pvc->GetCount(); i<iEnd; i++)
    {
        FdoPtr<FdoPropertyDefinition> pd = pvc->GetItem(i);
        if (wcscmp(pd->GetName(), ri[i].Name()) != 0)
            ALX_ERROR("FDO class definition does not match TableDef. Column values will mismatch!");
    }
}

void TableFdo::InitSelect()
{
    if (!_select)
    {
        _select = (FdoISelect*)_db->CreateCommand(FdoCommandType_Select);
        _select->SetFeatureClassName(_ri->Name());

        //Setting the property collection of the select command will ensure
        //we get the properties we want *in the order* we want
        //TODO: this stuff can be pre-created and cached
        //in case want to reuse the TableFdo
        //TODO: We don't have to populate this collection since the order in our
        //TableDef already matches that returned by a blank select command, however,
        //there is currently a bug in the SQLite provider which prevents us from
        //using property access by index if we don't populate this collection.
        FdoPtr<FdoIdentifierCollection> idc = _select->GetPropertyNames();
        for (int i=0; i<_ri->ColumnCount(); i++)
        {
            FdoPtr<FdoIdentifier> ident = FdoIdentifier::Create((*_ri)[i].Name());
            idc->Add(ident);
        }
    }

    _select->SetFilter((FdoFilter*)NULL);
}

void TableFdo::Query()
{
    if (_rdr)
        _rdr->Close();

    InitSelect();
    _rdr = _select->Execute();    

#ifdef _DEBUG
    FdoPtr<FdoClassDefinition> fc = _rdr->GetClassDefinition();
    CheckClassDef(fc, *_ri);
#endif
}

void TableFdo::QueryRange(__int64 minID, __int64 maxID)
{
    InitSelect();

    FdoPtr<FdoIdentifier> fdoid = FdoIdentifier::Create((*_ri)[_ri->IndexOfID()].Name());
    FdoPtr<FdoInt64Value> val1 = FdoInt64Value::Create(minID);
    FdoPtr<FdoInt64Value> val2 = FdoInt64Value::Create(maxID);
    FdoPtr<FdoComparisonCondition> cmp1 = FdoComparisonCondition::Create(fdoid, FdoComparisonOperations_GreaterThanOrEqualTo, val1);
    FdoPtr<FdoComparisonCondition> cmp2 = FdoComparisonCondition::Create(fdoid, FdoComparisonOperations_LessThan, val2);
    FdoPtr<FdoFilter> filter = FdoBinaryLogicalOperator::Create(cmp1, FdoBinaryLogicalOperations_And, cmp2);
    _select->SetFilter(filter);

    _rdr = _select->Execute();

#ifdef _DEBUG
    FdoPtr<FdoClassDefinition> fc = _rdr->GetClassDefinition();
    CheckClassDef(fc, *_ri);
#endif
}

void TableFdo::Query(const double bbox[4], FdoFilter* attributeFilter)
{
    //the FGF polygon for the spatial query
    FgfPolygon empty;
    empty.p[0] = bbox[0]; empty.p[1] = bbox[1];
	empty.p[2] = bbox[2]; empty.p[3] = bbox[1];
	empty.p[4] = bbox[2]; empty.p[5] = bbox[3];
	empty.p[6] = bbox[0]; empty.p[7] = bbox[3];
	empty.p[8] = bbox[0]; empty.p[9] = bbox[1];

	FdoPtr<FdoByteArray> fgf = FdoByteArray::Create((FdoByte*)&empty, sizeof(empty));

	//set up the spatial filter
	FdoPtr<FdoGeometryValue> gv = FdoGeometryValue::Create(fgf);
	FdoPtr<FdoSpatialCondition> cond = FdoSpatialCondition::Create(
		Def().Geom().Name(), FdoSpatialOperations_EnvelopeIntersects, gv);

    //TODO: combine with attribute filter

    InitSelect();

    _select->SetFilter(cond);

    _rdr = _select->Execute();
}

const Row* TableFdo::Query(__int64 id)
{
    if (!_idfilter)
    {
        FdoPtr<FdoIdentifier> name = FdoIdentifier::Create(_ri->ID().Name());
        _updid = FdoInt64Value::Create(-1); //this will be modified for every feature
        _idfilter = FdoComparisonCondition::Create(name, FdoComparisonOperations_EqualTo, _updid);
    }

    FdoPtr<FdoISelect> select = (FdoISelect*)_db->CreateCommand(FdoCommandType_Select);
    InitSelect();
    _updid->SetInt64(id);
    _select->SetFilter(_idfilter);
    
    _rdr = _select->Execute();
    return Next();
}

void TableFdo::EndQuery()
{
    if (_rdr)
    {
        _rdr->Close();
        _rdr = NULL;
    }
}

void TableFdo::Query(FdoFilter* f)
{
    FdoPtr<FdoISelect> select = (FdoISelect*)_db->CreateCommand(FdoCommandType_Select);
    InitSelect();
    _select->SetFilter(f);
    _rdr = _select->Execute();
}

void TableFdo::GetExtent(double bbox[4])
{
    FdoPtr<FdoISelectAggregates> agg = (FdoISelectAggregates*)_db->CreateCommand(FdoCommandType_SelectAggregates);
    agg->SetFeatureClassName(Def().Name());
    FdoPtr<FdoIdentifierCollection> props = agg->GetPropertyNames();

    std::wstring exps = std::wstring(L"SpatialExtents(") + Def().Geom().Name() + L")";
    FdoPtr<FdoExpression> exp = FdoExpression::Parse(exps.c_str());
    FdoPtr<FdoComputedIdentifier> se = FdoComputedIdentifier::Create(L"Extents", exp);
    props->Add(se);

    std::wstring expc = std::wstring(L"Count()");
    exp = FdoExpression::Parse(expc.c_str());
    se = FdoComputedIdentifier::Create(L"TheCount", exp);
    props->Add(se);

    FdoPtr<FdoIDataReader> extrdr = agg->Execute();
    extrdr->ReadNext();

    FdoPtr<FdoByteArray> ba = extrdr->GetGeometry(L"Extents");

    FgfPolygon* poly = (FgfPolygon*)ba->GetData();

    bbox[0] = poly->p[0];
    bbox[1] = poly->p[1];
    bbox[2] = poly->p[4];
    bbox[3] = poly->p[5];

    extrdr->Close();
}

__int64 TableFdo::GetRowCount()
{
    FdoPtr<FdoISelectAggregates> agg = (FdoISelectAggregates*)_db->CreateCommand(FdoCommandType_SelectAggregates);
    agg->SetFeatureClassName(Def().Name());
    FdoPtr<FdoIdentifierCollection> props = agg->GetPropertyNames();

    std::wstring expc = std::wstring(L"Count()");
    FdoPtr<FdoExpression> exp = FdoExpression::Parse(expc.c_str());
    FdoPtr<FdoComputedIdentifier> se = FdoComputedIdentifier::Create(L"TheCount", exp);
    props->Add(se);

    FdoPtr<FdoIDataReader> extrdr = agg->Execute();
    extrdr->ReadNext();

    __int64 ret = extrdr->GetInt64(L"TheCount");

    extrdr->Close();

    return ret;
}

const Row* TableFdo::Next()
{
    FdoIFeatureReader* rdr = _rdr.p;

    if (!rdr)
        return NULL;

    try
    {
        if (rdr->ReadNext())
        {
            const TableDef& ri = *_ri;
            Row& row = *_row;
            for (int i=0; i<ri.ColumnCount(); i++)
            {
                if (rdr->IsNull(i))
                {
                    row[i].Reset();
                    continue;
                }

                int type = ri[i].DataType();

                switch (type)
                {
                case FdoDataType_BLOB:
                    {
                        FdoPtr<FdoLOBValue> lob = rdr->GetLOB(i);
                        FdoPtr<FdoByteArray> ba = lob->GetData();
                        row[i].Reset((void*)ba->GetData(), ba->GetCount()); 
                    }
                    break;
                case FdoDataType_Boolean: row[i].Reset(rdr->GetBoolean(i)); break;
                case FdoDataType_Byte: row[i].Reset(rdr->GetByte(i)); break;
                case FdoDataType_CLOB:
                    {
                        FdoPtr<FdoLOBValue> lob = rdr->GetLOB(i);
                        FdoPtr<FdoByteArray> ba = lob->GetData();
                        row[i].Reset((void*)ba->GetData(), ba->GetCount());
                    }
                    break;
                case FdoDataType_DateTime:
                    {
                        FdoDateTime dt = rdr->GetDateTime(i);
                        wchar_t str[64];
                        if (DateToString(dt, str, 64))
                        {
                            row[i].Reset(str);
                        }
                        else
                        {
                            row[i].Reset();
                        }
                    }
                    break;
                case FdoDataType_Decimal: row[i].Reset(rdr->GetDouble(i)); break;
                case FdoDataType_Double:  row[i].Reset(rdr->GetDouble(i)); break;
                case FdoDataType_Int16:   row[i].Reset(rdr->GetInt16(i));  break;
                case FdoDataType_Int32:   row[i].Reset(rdr->GetInt32(i));  break;
                case FdoDataType_Int64:   row[i].Reset(rdr->GetInt64(i));  break;
                case FdoDataType_Single:  row[i].Reset(rdr->GetSingle(i)); break;
                case FdoDataType_String:  row[i].Reset(rdr->GetString(i), false); break;
                case ColumnDef::DataType_Geom: //our special value indicating geom property
                    {
                        int len = 0;
                        const unsigned char* fgf = rdr->GetGeometry(i, &len);
                        row[i].Reset((void*)fgf, len, false);
                    }
                    break;
                default:
                    break;
                }
            }

            return _row;
        }
        else
        {
            Row& row = *_row;
            for (int i=0; i<row.Def().ColumnCount(); i++)
                row[i].Invalidate();

            return NULL;
        }
    }
    catch (FdoException* e)
    {
        ALX_ERROR(e->GetExceptionMessage());
        e->Release();
        return NULL;
    }
}


//=============================================================================
//=============================================================================
// DatabaseFdo
//=============================================================================
//=============================================================================

DatabaseFdo::DatabaseFdo()
{
    _db = NULL;
    _schema = NULL;
    _tables = NULL;
}

DatabaseFdo::~DatabaseFdo()
{
    Disconnect();
}

bool DatabaseFdo::Connect(FdoIConnection* db)
{
    Disconnect();

    _db = FDO_SAFE_ADDREF(db);
    _connStr = db->GetConnectionString();

    InitSchema();

    return true;
}

bool DatabaseFdo::Connect(const wchar_t* connStr, bool readOnly)
{
    Disconnect();

    _db = GetFdoCon(connStr, true);

    if (!_db)
        return false;

    _connStr = connStr;
    InitSchema();
    return true;
}

bool DatabaseFdo::Create(const wchar_t* filePath)
{
    FdoPtr<FdoIConnection> con;
    try
    {
        con = GetFdoCon(filePath, false);

        if (!con)
            return false;

        //create the destination database file
        FdoPtr<FdoICreateDataStore> cds = (FdoICreateDataStore*)con->CreateCommand(FdoCommandType_CreateDataStore);
        FdoPtr<FdoIDataStorePropertyDictionary> dspd = cds->GetDataStoreProperties();

        //TODO: needs different logic for SHP files
        dspd->SetProperty(L"File", filePath);
        dspd->SetProperty(L"UseFdoMetadata", L"TRUE");

        cds->Execute();
    }
    catch (FdoException* e)
    {
        fprintf (stderr, "Failed to create database, because: %ls\n", e->GetExceptionMessage());
        e->Release();
        return false;
    }

    return true;
}

void DatabaseFdo::Disconnect()
{
    _connStr.clear();
    ClearSchema();
    if (_db)
        _db->Close();
    FDO_SAFE_RELEASE(_db);
    _db = NULL;
}


void DatabaseFdo::InitSchema()
{
    //get the coord sys list first
    FdoPtr<FdoIGetSpatialContexts> sc = (FdoIGetSpatialContexts*)_db->CreateCommand(FdoCommandType_GetSpatialContexts);
    FdoPtr<FdoISpatialContextReader> rdr = sc->Execute();

    //accumulate in a temp vector 
    std::vector<CoordSys> cslist;
    while (rdr->ReadNext())
    {
        cslist.push_back(CoordSys(rdr->GetName(), rdr->GetCoordinateSystem(), rdr->GetCoordinateSystemWkt()));
    }

    //DescribeSchema usually returns a cloned schema.
    //The FDO schema cloning code changes the order of the properties
    //in the class' property collection, which is undesirable for performance reasons.
    //Therefore, here we create a blank select command and get its class definition,
    //which (for providers we really care about) does not pass through the bad cloning code
    FdoPtr<FdoIDescribeSchema> ds = (FdoIDescribeSchema*)_db->CreateCommand(FdoCommandType_DescribeSchema);
    FdoPtr<FdoFeatureSchemaCollection> fsc = ds->Execute();

    if (fsc->GetCount() > 0)
    {
        FdoPtr<FdoFeatureSchema> schema = fsc->GetItem(0);
        FdoPtr<FdoClassCollection> classes = schema->GetClasses();

        int count = classes->GetCount();

        if (count)
        {
            TableDef* rowdefs = new TableDef[count];
            CoordSysDef* csdef = cslist.size() ? new CoordSysDef(&cslist[0], cslist.size()) : NULL;

            for (int i=0, iEnd = classes->GetCount(); i<iEnd; i++)
            {
                FdoPtr<FdoClassDefinition> fc = classes->GetItem(i);
                FdoPtr<FdoISelect> select = (FdoISelect*)_db->CreateCommand(FdoCommandType_Select);
                select->SetFeatureClassName(fc->GetName());
                FdoPtr<FdoIFeatureReader> rdr = select->Execute();
                FdoPtr<FdoClassDefinition> myfc = rdr->GetClassDefinition();

                FdoClassToTableDef(rowdefs[i], myfc, *csdef);
            }

            //here, the DatabaseDef takes ownership of these pointers
            //TODO: the DatabaseDef should make in internal copy of these
            //to avoid potential cross-DLL heap issues
            _schema = new DatabaseDef(schema->GetName(), rowdefs, count, csdef);

            //create the physical tables, based on each table def
            _tables = new Table*[_schema->Count()];
            for (int i=0; i<_schema->Count(); i++)
            {
                _tables[i] = new TableFdo(_db, &(*_schema)[i]);
            }
        }
    }
       

}


bool DatabaseFdo::SetSchema(const DatabaseDef& def)
{
    //create the spatial contexts
    SetCoordSys(def.CSDef());

    FdoPtr<FdoIApplySchema> as = (FdoIApplySchema*)_db->CreateCommand(FdoCommandType_ApplySchema);

    FdoPtr<FdoFeatureSchema> schema = FdoFeatureSchema::Create(def.Name(), NULL);
    FdoPtr<FdoClassCollection> classes = schema->GetClasses();

    for (int i=0; i<def.Count(); i++)
    {
        FdoPtr<FdoClassDefinition> fc = TableDefToFdoClass(def[i]);
        classes->Add(fc);
    }

    as->SetFeatureSchema(schema);

    try
    {
        as->Execute();

        //reset the cached schema
        ClearSchema();
        InitSchema();
    }
    catch (FdoException* e)
    {
        fprintf (stderr, "Failed to apply schema, because: %ls\n", e->GetExceptionMessage());
        e->Release();
        return false;
    }
    
    return true;
}

//Transfer spatial context definitions from the given CoordSysDef into 
//the currently connected database
bool DatabaseFdo::SetCoordSys(const CoordSysDef& def)
{
    try
    {
        for (int i=0; i<def.Count(); i++)
        {
            FdoPtr<FdoICreateSpatialContext> spc = (FdoICreateSpatialContext*)_db->CreateCommand(FdoCommandType_CreateSpatialContext);
            
            spc->SetName(def[i].Name());
            spc->SetCoordinateSystem(def[i].Code());
            spc->SetCoordinateSystemWkt(def[i].Wkt());

            spc->Execute();
        }
    }
    catch (FdoException* e)
    {
        fprintf (stderr, "Failed to create spatial context, because: %ls\n", e->GetExceptionMessage());
        e->Release();
        return false;
    }

    return true;
}



}



#endif