//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#include "stdafx.h"

#if !EXCLUDE_OGR

#include "OgrWrap.h"

#include <ogrsf_frmts.h>

namespace fdo
{

//===========================================================================
//
// FGF - WKB conversion
//TODO: Could use OGR built-in conversion, since it has FGF support now
//
//===========================================================================

class OgrBinaryReader
{
public:
    OgrBinaryReader(const unsigned char* stream)
        : m_stream((unsigned char*)stream)
    {
    }

    inline unsigned char ReadByte()
    {
        return *m_stream++;
    }

    inline int ReadInt()
    {
        int ret = *(int*)m_stream;
        m_stream += sizeof(int);
        return ret;
    }

    inline double ReadDouble()
    {
        double ret = *(double*)m_stream;
        m_stream += sizeof(double);
        return ret;
    }

    unsigned char* m_stream;
};


class OgrBinaryWriter
{
public:
    OgrBinaryWriter(unsigned char* stream /*, int len*/)
        : m_stream(stream),
        m_begin(stream)
    {
    }

    inline void WriteByte(unsigned char b)
    {
        *m_stream++ = b;
    }

    inline void WriteInt(int i)
    {
        *((int*)m_stream) = i;
        m_stream += sizeof(int);
    }

    inline void WriteDouble(double d)
    {
        *((double*)m_stream) = d;
        m_stream += sizeof(double);
    }

    inline int GetLength() 
    {
        return (int)(m_stream - m_begin);
    }

    unsigned char* m_begin;
    unsigned char* m_stream;
};


//For reference, duplicated from ogr_geometry.h
///**
// * List of well known binary geometry types.  These are used within the BLOBs
// * but are also returned from OGRGeometry::getGeometryType() to identify the
// * type of a geometry object.
// */
//
//typedef enum 
//{
//    wkbUnknown = 0,             /* non-standard */
//    wkbPoint = 1,               /* rest are standard WKB type codes */
//    wkbLineString = 2,
//    wkbPolygon = 3,
//    wkbMultiPoint = 4,
//    wkbMultiLineString = 5,
//    wkbMultiPolygon = 6,
//    wkbGeometryCollection = 7,
//    wkbNone = 100,              /* non-standard, for pure attribute records */
//    wkbLinearRing = 101,        /* non-standard, just for createGeometry() */
//    wkbPoint25D = 0x80000001,   /* 2.5D extensions as per 99-402 */
//    wkbLineString25D = 0x80000002,
//    wkbPolygon25D = 0x80000003,
//    wkbMultiPoint25D = 0x80000004,
//    wkbMultiLineString25D = 0x80000005,
//    wkbMultiPolygon25D = 0x80000006,
//    wkbGeometryCollection25D = 0x80000007
//} OGRwkbGeometryType;
//
//#define wkb25DBit 0x80000000
//#define wkbFlatten(x)  ((OGRwkbGeometryType) ((x) & (~wkb25DBit)))
//
//#define ogrZMarker 0x21125711
//
//const char CPL_DLL * OGRGeometryTypeToName( OGRwkbGeometryType eType );
//
//typedef enum 
//{
//    wkbXDR = 0,         /* MSB/Sun/Motoroloa: Most Significant Byte First   */
//    wkbNDR = 1          /* LSB/Intel/Vax: Least Significant Byte First      */
//} OGRwkbByteOrder;


static int Wkb2Fgf(const unsigned char* wkb, unsigned char* fgf)
{
    OgrBinaryReader src(wkb);
    OgrBinaryWriter dst(fgf);

    src.ReadByte(); //skip byte order -- FGF is always NDR

    // the geometry type
    int geom_type = src.ReadInt();

    // the coordinate type
    int skip = (geom_type < 0) ? 1 : 0; //0=XY, 1=XYZ or XYM, 2 = XYZM
    geom_type = geom_type & 0x7FFFFFFF; //now convert to FGF geom type

    dst.WriteInt(geom_type); //strip out WKB 2.5D indicator (FDO doesn't use that)

    bool is_multi = (geom_type == FdoGeometryType_MultiLineString)
        || (geom_type == FdoGeometryType_MultiPolygon
        || (geom_type == FdoGeometryType_MultiPoint));

    int num_geoms = 1;

    //in case of multipolygon or multilinestring or multipoint,
    //read poly or linestring count
    if (is_multi) 
    {
        num_geoms = src.ReadInt();
        dst.WriteInt(num_geoms);
    }

    for (int q=0; q<num_geoms; q++)
    {
        if (is_multi)
        {
            //skip byte order
            src.ReadByte();

            //geom type
            geom_type = src.ReadInt();
            skip = geom_type < 0 ? 1 : 0; //is it 2.5D
            geom_type &= 0x7FFFFFFF;
            dst.WriteInt(geom_type);
        }

        //read cordinate type
        int dim = skip ? FdoDimensionality_XY | FdoDimensionality_Z : FdoDimensionality_XY;
        dst.WriteInt(dim);

        // the number of contours in current polygon/linestring
        int contour_count = 1; //for linestrings

        if ((geom_type == FdoGeometryType_Polygon)
            || (geom_type == FdoGeometryType_MultiPolygon))
        {
            contour_count = src.ReadInt();
            dst.WriteInt(contour_count);
        }

        for (int i=0; i<contour_count; i++)
        {
            int point_count = 1;

            //point geoms do not have a point count, since
            //each piece is just one point each
            if ((geom_type != FdoGeometryType_MultiPoint)
                && (geom_type != FdoGeometryType_Point))
            {
                point_count = src.ReadInt();
                dst.WriteInt(point_count);
            }

            int numd = point_count * (skip + 2);
            for (int j=0; j<numd; j++) dst.WriteDouble(src.ReadDouble());
        }
    }

    return dst.GetLength();
}


static int Fgf2Wkb(const unsigned char* fgf, unsigned char* wkb)
{
    OgrBinaryReader src(fgf);
    OgrBinaryWriter dst((unsigned char*)wkb);

    dst.WriteByte(1); //NDR

    // the geometry type
    int geom_type = src.ReadInt();

    //we will need to do a second pass to set the 3D flag on this once we read the FDO dimensionality flag
    dst.WriteInt(geom_type); 

    //The FDO constants match WKB, so we don't really need to use the OGR wkb defines
    bool is_multi = (geom_type == FdoGeometryType_MultiLineString)
        || (geom_type == FdoGeometryType_MultiPolygon
        || (geom_type == FdoGeometryType_MultiPoint));

    int num_geoms = 1;

    //in case of multipolygon or multilinestring or multipoint,
    //read poly or linestring count
    if (is_multi) 
    {
        num_geoms = src.ReadInt();
        dst.WriteInt(num_geoms);
    }

    for (int q=0; q<num_geoms; q++)
    {
        if (is_multi)
        {
            //set byte order
            dst.WriteByte(1);

            //geom type
            geom_type = src.ReadInt();
            dst.WriteInt(geom_type);
        }

        //read cordinate type
        int dim = src.ReadInt();
        int skip = ((dim & FdoDimensionality_Z) != 0) ? 1 : 0;

        if (skip)
        {
            //roll back and set the geom_type to wkb 2.5D
            //TODO: do we need to set both the geom type for the parent multi geom
            //and the child geom? For example in the case of a multipolygon, do we need to 
            //set it for both the current polygon and the parent multipolygon?
            *(int*)(dst.m_begin+1) = geom_type | 0x80000000; 
        }
        
        // the number of contours in current polygon/linestring
        int contour_count = 1; //for linestrings

        if ((geom_type == FdoGeometryType_Polygon)
            || (geom_type == FdoGeometryType_MultiPolygon))
        {
            contour_count = src.ReadInt();
            dst.WriteInt(contour_count);
        }

        for (int i=0; i<contour_count; i++)
        {
            int point_count = 1;

            //point geoms do not have a point count, since
            //each piece is just one point each
            if ((geom_type != FdoGeometryType_MultiPoint)
                && (geom_type != FdoGeometryType_Point))
            {
                point_count = src.ReadInt();
                dst.WriteInt(point_count);
            }

            int numd = point_count * (skip + 2);
            for (int j=0; j<numd; j++) dst.WriteDouble(src.ReadDouble());
        }
    }

    return dst.GetLength();
}




//===========================================================================
//
//   DatabaseOgr
//
//===========================================================================

int ogr_registered = 0;


DatabaseOgr::DatabaseOgr()
{
    //TODO: probably needs run-once protection of some sort
    if (!ogr_registered)
    {
        OGRRegisterAll();
        ogr_registered = 1;
    }

    _db = NULL;
}

DatabaseOgr::~DatabaseOgr()
{
    Disconnect();
}

bool DatabaseOgr::Connect(const wchar_t* connStr, bool readOnly)
{
    Disconnect();

    std::string ogrStr = W2A_SLOW(connStr);

    _db = OGRSFDriverRegistrar::Open(ogrStr.c_str(), !readOnly);

    if (!_db)
        return false;

    _connStr = connStr;
    InitSchema();
    return true;
}

void DatabaseOgr::Disconnect()
{
    _connStr.clear();
    ClearSchema();

    if (_db)
    {
        OGRDataSource::DestroyDataSource(_db);
        _db = NULL;
    }
}

bool DatabaseOgr::SetSchema(const DatabaseDef& def)
{
    ALX_ERROR("Ogr::SetSchema not implemented\n");
    return false;
}

static void OgrLayerToTableDef(TableDef& tdef, OGRLayer* layer, const CoordSys* cs)
{
    OGRFeatureDefn* fdef = layer->GetLayerDefn();

    //ID and geom will use fixed index for OGR,
    //since they are separate from the other fields

    int idindex = 0;
    int geomindex = 1;
    std::wstring idname;
    std::wstring geomname;

    //id property name
    const char* tmp = layer->GetFIDColumn();
    if (tmp && *tmp)
    {
        idname = A2W_SLOW(tmp);
    }
    else
    {
        idname = L"OGR_FID"; //TODO: should really make sure this is unique (no other columns named like this)
    }
    
    //geom property name
    tmp = layer->GetGeometryColumn();
    if (tmp && *tmp)
    {
        geomname = A2W_SLOW(tmp);
    }
    else
    {
        geomname = L"OGR_GEOMETRY"; //TODO: should really make sure this is unique (no other columns named like this)
    }

    //build the name-index map and also collect info about each column
    int nprops = fdef->GetFieldCount();
    ColumnDef* cols = new ColumnDef[2 + nprops];

    cols[0].Initialize(idname.c_str(), FdoDataType_Int64, NULL);
    cols[1].Initialize(geomname.c_str(), (FdoDataType)ColumnDef::DataType_Geom, cs);

    for (int i=0; i<nprops; i++)
    {
        OGRFieldDefn* pd = fdef->GetFieldDefn(i);

        OGRFieldType ft = pd->GetType();
        FdoDataType dt = (FdoDataType)ColumnDef::DataType_Unknown;

        switch (ft)
        {
        case OFTInteger : dt = FdoDataType_Int32; break;
        case OFTIntegerList : break;
        case OFTReal : dt = FdoDataType_Double; break;
        case OFTRealList : break;
        case OFTString : dt = FdoDataType_String; break;
        case OFTStringList : break;
        case OFTWideString : dt = FdoDataType_String; break;
        case OFTWideStringList : break;
        case OFTBinary : dt = FdoDataType_BLOB; break;
        case OFTDate : 
        case OFTTime : 
        case OFTDateTime : dt = FdoDataType_DateTime; break;
        }

        cols[i+2].Initialize(A2W_SLOW(pd->GetNameRef()).c_str(), dt, NULL);
       
    }

    tdef.Initialize(A2W_SLOW(fdef->GetName()).c_str(), cols, 2+nprops, idindex, geomindex);
  
}

void DatabaseOgr::InitSchema()
{
    int count = _db->GetLayerCount();
    
    if (count)
    {
        TableDef* rowdefs = new TableDef[count];
        CoordSysDef* csdef = NULL;

        std::vector<CoordSys> cslist;

        for (int i=0; i<count; i++)
        {
            OGRLayer* lr = _db->GetLayer(i);
            OGRSpatialReference* srs = lr->GetSpatialRef();

            if (srs)
            {
                char* wkt = NULL;
                srs->exportToWkt(&wkt);

                if (wkt)
                {
                    std::wstring wwkt = A2W_SLOW(wkt);
                    
                    //make a name for the coord sys, simply based
                    //on the layer name
                    std::wstring wname = A2W_SLOW(lr->GetLayerDefn()->GetName());
                    wname += L"_srs";

                    cslist.push_back(CoordSys(wname.c_str(), NULL, wwkt.c_str()));
                    
                    OGRFree(wkt);
                }
            }
            else
            {
                cslist.push_back(CoordSys(L"", L"", L""));
            }
        }

        csdef = new CoordSysDef(&cslist[0], cslist.size());

        for (int i=0; i<count; i++)
        {
            OGRLayer* lr = _db->GetLayer(i);
            OgrLayerToTableDef(rowdefs[i], lr, &(*csdef)[i]);
        }

        _schema = new DatabaseDef(L"OGRDefaultSchema" /*A2W_SLOW(_db->GetName()).c_str()*/, rowdefs, count, csdef);

        //create the physical tables, based on each table def
        _tables = new Table*[_schema->Count()];
        for (int i=0; i<_schema->Count(); i++)
        {
            _tables[i] = new TableOgr(_db->GetLayer(i), &(*_schema)[i]);
        }
    }
}






//===========================================================================
//
//   TableOgr
//
//===========================================================================



TableOgr::TableOgr(OGRLayer* layer, const TableDef* td)
{
    _layer = layer;
    _ri = td;
    _row = new Row(td);
    _fgfbuf = NULL;
    _fgflen = 0;
    _wkbbuf = NULL;
    _wkblen = 0;
}

TableOgr::~TableOgr()
{
    delete _row;
    delete [] _fgfbuf;
    delete [] _wkbbuf;
}

bool TableOgr::Insert(const Row& row, __int64* newfid, bool skipID)
{
    OGRFeature* f = OGRFeature::CreateFeature(_layer->GetLayerDefn());
    RowToFeature(f, row);
    
    if (skipID)
        f->SetFID(OGRNullFID);

    OGRErr res = _layer->CreateFeature(f);

    //if caller requested we return the id of the new feature
    //set it
    if (newfid)
    {
        if (res == OGRERR_NONE)
        {
            long l = f->GetFID();
            *newfid = l == OGRNullFID ? INVALID_FID : l;
        }
        else
            *newfid = INVALID_FID;
    }

    return res == OGRERR_NONE;
}

bool TableOgr::Update(const Row& row)
{
    OGRFeature* f = OGRFeature::CreateFeature(_layer->GetLayerDefn());
    RowToFeature(f, row);
    
    OGRErr res = _layer->SetFeature(f);

    return res == OGRERR_NONE;
}


bool TableOgr::Delete(__int64 fid)
{
    OGRErr res = _layer->DeleteFeature(fid);
    return res == OGRERR_NONE;
}

//Query operations
void TableOgr::Query()
{
    _layer->SetAttributeFilter(NULL);
    _layer->SetSpatialFilter(NULL);
    _layer->ResetReading();
}

void TableOgr::QueryRange(__int64 minID, __int64 maxID)
{
    throw;
}

void TableOgr::QueryRandomAccess() 
{
    throw;
} 

void TableOgr::Query(const double bbox[4], FdoFilter* attributeFilter)
{
    if (attributeFilter)
    {
        const wchar_t* where_clause = attributeFilter->ToString();
        _layer->SetAttributeFilter(W2A_SLOW(where_clause).c_str());
    }
    else
        _layer->SetAttributeFilter(NULL);

    _layer->SetSpatialFilterRect(bbox[0], bbox[1], bbox[2], bbox[3]);
    _layer->ResetReading();
}

const Row* TableOgr::Query(__int64 id)
{
    OGRFeature* f = _layer->GetFeature(id);

    if (f)
    {
        Row* r = _row;
        FeatureToRow(*r, f);
        OGRFeature::DestroyFeature(f);
        return r;
    }
    else
    {
        return NULL;
    }
}

void TableOgr::Query(FdoFilter*)
{
}

void TableOgr::EndQuery()
{
    _layer->ResetReading();
}

void TableOgr::GetExtent(double bbox[4])
{
    OGREnvelope ext;
    _layer->GetExtent(&ext);
    bbox[0] = ext.MinX;
    bbox[1] = ext.MinY;
    bbox[2] = ext.MaxX;
    bbox[3] = ext.MaxY;
}

__int64 TableOgr::GetRowCount()
{
    return _layer->GetFeatureCount();
}

void TableOgr::GeomToValue(Value& v, OGRGeometry* geom)
{
    //set geometry
    int len = geom->WkbSize();
    if (_wkblen < len)
    {
        delete [] _wkbbuf;
        _wkblen = len;
        _wkbbuf = new unsigned char[len];
    }
    int fgflen = 2*len;
    if (_fgflen < fgflen)
    {
        delete [] _fgfbuf;
        _fgflen = fgflen;
        _fgfbuf = new unsigned char[fgflen];
    }

    //the conversion call returns the actual length
    geom->exportToWkb(wkbNDR, _wkbbuf);
    fgflen = Wkb2Fgf(_wkbbuf, _fgfbuf);
    v.Reset(_fgfbuf, fgflen, false);
}


OGRGeometry* TableOgr::ValueToGeom(const Value& v)
{
    int len = v.BlobLen();
    const unsigned char* fgf = (const unsigned char*)v.AsBlob();

    if (!fgf)
    {
        return NULL;
    }

    if (_wkblen < len)
    {
        delete [] _wkbbuf;
        _wkblen = len;
        _wkbbuf = new unsigned char[len];
    }

    int wkblen = Fgf2Wkb(fgf, _wkbbuf);
    OGRGeometry* geom = NULL;    
    OGRGeometryFactory::createFromWkb(_wkbbuf, _layer->GetSpatialRef(), &geom); 
    return geom;
}


void TableOgr::RowToFeature(OGRFeature* f, const Row& r)
{
    int ncols = r.Def().ColumnCount();

    f->SetFID(r[0].AsInt());

    OGRGeometry* geom = ValueToGeom(r[1]);
    f->SetGeometryDirectly(geom);

    for (int i=2; i<ncols; i++)
    {
        int ogri = i-2;

        if (r[i].IsNull())
            f->UnsetField(ogri);
        
        int t = Def()[i].DataType();

        switch (t)
        {
        case FdoDataType_BLOB:
            {
                f->SetField(ogri, r[i].BlobLen(), (GByte*)r[i].AsBlob());
            }
            break;
        case FdoDataType_DateTime:
            {
                FdoDateTime dt;
                if (DateFromString(dt, r[i].AsString()))
                {
                    f->SetField(ogri, dt.year, dt.month, dt.day, dt.hour, dt.minute, (int)dt.seconds, 0);
                }
                else
                {
                    f->UnsetField(ogri);
                }
            }
            break;
        case FdoDataType_Double:  f->SetField(ogri, r[i].AsReal()); break;
        case FdoDataType_Int32:   f->SetField(ogri, (int)r[i].AsInt());  break;
        case FdoDataType_Int64:   f->SetField(ogri, (int)r[i].AsInt());  break;
        case FdoDataType_String: 
            {
                const wchar_t* str = r[i].AsString();
                size_t wlen = wcslen(str);
                size_t clen = 4 * wlen+1;
                char* mbfc = (char*)alloca(clen);
                W2A_FAST(mbfc, clen, str, wlen);
                f->SetField(ogri, mbfc);
            }
            break;
        default:
            break;
        }

    }
}

void TableOgr::FeatureToRow(Row& r, OGRFeature* f)
{
    int count = r.Def().ColumnCount();

    //set ID
    r[0].Reset((__int64)f->GetFID());

    //convert the geometry
    GeomToValue(r[1], f->GetGeometryRef());

    //Process remaining columns
    for (int i=2; i<count; i++)
    {
        int ogri = i-2;

        if (!f->IsFieldSet(ogri))
        {
            r[i].Reset();
            continue;
        }

        int t = Def()[i].DataType();

        switch (t)
        {
        case FdoDataType_BLOB:
            {
                int len = 0;
                GByte* blob = f->GetFieldAsBinary(ogri, &len);
                r[i].Reset(blob, len, true);
            }
            break;
        case FdoDataType_DateTime:
            {
                int year, month, day, hour, minute, second, tz;

                int res = f->GetFieldAsDateTime(ogri, &year, &month, &day, &hour, &minute, &second, &tz);

                if (res)
                {
                    FdoDateTime dt(year, month, day, hour, minute, second);

                    wchar_t str[64];
                    if (DateToString(dt, str, 64))
                    {
                        r[i].Reset(str);
                    }
                    else
                    {
                        r[i].Reset();
                    }
                }
                else
                {
                    r[i].Reset();
                }
            }
            break;
        case FdoDataType_Double:  r[i].Reset(f->GetFieldAsDouble(ogri)); break;
        case FdoDataType_Int32:   r[i].Reset(f->GetFieldAsInteger(ogri));  break;
        case FdoDataType_Int64:   r[i].Reset(f->GetFieldAsInteger(ogri));  break;
        case FdoDataType_String: 
            {
                const char* mbs = f->GetFieldAsString(ogri);
                int len = (int)strlen(mbs);
                wchar_t* tmp = (wchar_t*)alloca(sizeof(wchar_t)*(len+1));
                A2W_FAST(tmp, len+1, mbs, len);
                r[i].Reset(tmp, true);
            }
            break;
        default:
            break;
        }
    }
}


const Row* TableOgr::Next()
{
    OGRFeature* f;
    if (f = _layer->GetNextFeature())
    {
        Row* r = _row;
        FeatureToRow(*r, f);
        OGRFeature::DestroyFeature(f);
        return r;
    }

    return NULL;
}


}


#endif