//
//  Copyright (C) 2010 by Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#pragma once

#include "FdoWrapDll.h"

#ifdef _MSC_VER
  #define ALGNW __declspec(align(16))
  #define ALGNL
#else
  #define ALGNW
  #define ALGNL __attribute__((aligned(16)))
#endif

#include <assert.h>

namespace fdo
{

//Data types native to the wrapper Value type
//Important -- should not exceed 15 since we 
//only hae 4 bits to store this!!!
enum ValueType
{
    vtUndefined = 0,
    vtNull = 1,
    vtBool = 2,
    vtInteger = 3,
    vtReal = 4,
    vtString = 5,
    vtBlob = 6,
    vtStringNotOwn = vtString + 0x8, //String using externally managed memory pointer
    vtBlobNotOwn = vtBlob + 0x8, //Blob using externally managed memory pointer
};


//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

//===========================================================================
//===========================================================================
//    Value -- encapsulates the value of a single column in a row
//    provides automatic type conversions
//===========================================================================
//===========================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

#pragma pack(push, 1)
class FDOWRAP_API ALGNW Value
{

private:

    union
    {
        struct
        {
            int type;
            int size; //for blobs
        };
        mutable const wchar_t* ptostr; //cached string representation for simple types
                               //overlaps with type and size, but for simple types size
                               //is not used, while type fits in the last 3 bits of ptostr,
                               //which we will strategically allocate on a 16 byte boundary
                               //so that the last 4 bits are 0 and available for us to cram things into.
    };

    union
    {
        bool     bval;
        wchar_t* pstr;
        void*    pblob;
        __int64  ival;
        double   rval;
    };

public:

    Value()                   { SetNull(); }
    Value(__int64 i)          { SetInt(i); }
    Value(double d)           { SetDouble(d); }
    Value(const wchar_t* str) { SetString(str, true); }
    Value(const void* blob, int len) { SetBlob(blob, len, true); }

    ~Value();

    int Type() const { return type & 0xf; }
    bool IsNull() const { return (Type() == vtNull); }
    bool IsValid() const { return (Type() != vtUndefined); }

    bool AsBool() const
    {
        int t = Type();
        switch(t)
        {
        case vtBool: return bval;
        case vtInteger: return !!ival;
        case vtStringNotOwn:
        case vtString: return _wcsicmp(pstr, L"true") == 0;
        default: throw; //TODO: something clever that doesn't crash here
        }
    }
    operator bool() const { return AsBool(); }

    __int64 AsInt() const
    {
        int t = Type();
        switch(t)
        {
        case vtInteger: return ival;
        case vtReal: return (__int64)rval;
        case vtStringNotOwn:
        case vtString: return _wtoi64(pstr);
        default: throw; //TODO: something clever that doesn't crash here
        }
    }
    operator __int64() const { return AsInt(); }

    double AsReal() const
    {
        int t = Type();
        switch(t)
        {
        case vtReal: return rval;
        case vtInteger: return (double)ival;
        case vtStringNotOwn:
        case vtString: return _wtof(pstr); //TODO: must use C locale always for the decimal sep
        default: throw; //TODO: something clever that doesn't crash here
        }
    }
    operator double() const { return AsReal(); }

    const wchar_t* AsString() const
    {
        int t = Type();
        switch(t)
        {        
        case vtStringNotOwn:
        case vtString: return pstr;
        case vtNull: return NULL;
        case vtBool: 
        case vtInteger: 
        case vtReal: return CachedStringPtr();
        default: return NULL; //TODO: something clever here
        }
    }
    operator const wchar_t*() const { return AsString(); }

    const unsigned char* AsBlob() const
    {
        int t = Type();
        switch(t)
        {       
        case vtBlobNotOwn:
        case vtBlob: return (const unsigned char*)pblob;
        default: return NULL;
        }
    }

    int BlobLen() const { return size; }

    void Reset(__int64 i)
    {
        FreeMem();
        SetInt(i);
    }

    void Reset(int i)
    {
        FreeMem();
        SetInt(i);
    }

    void Reset(bool b)
    {
        FreeMem();
        SetBool(b);
    }

    void Reset(double d)
    {
        FreeMem();
        SetDouble(d);
    }

    void Reset(float d)
    {
        FreeMem();
        SetDouble(d);
    }

    
    void Reset(const wchar_t* str, bool own = true)
    {
        FreeMem();
        SetString(str, own);
    }
    
    void Reset(const void* blob, int len, bool own = true)
    {
        FreeMem();
        SetBlob(blob, len, own);
    }
    
    //resets the value to NULL  
    void Reset()
    {
        FreeMem();
        SetNull();
    }

    //invalidates the contents of the Value
    void Invalidate()
    {
        FreeMem();
        SetUndef();
    }

private:
    void FreeMem();
    void CacheString();
    const wchar_t* CachedStringPtr() const;

    //NOTE: These set functions do not free any existing allocated memory
    //for blobs and strings! FreeMem() must be called before calling these.
    //That is why they are private.
    void SetString(const wchar_t* str, bool own);
    void SetBlob(const void* blob, int len, bool own);

    void SetUndef()
    {
        type = vtUndefined;
        size = 0;
        ival = 0; //should set all the union members to 0;
    }

    void SetBool(bool b)
    {
        type = vtBool;
        size = 0;
        bval = b;
    }

    void SetNull()
    {
        type = vtNull;
        size = 0;
        ival = 0; 
    }

    void SetInt(__int64 i)
    {
        type = vtInteger;
        ival = i;
        size = 0;
    }

    void SetDouble(double d)
    {
        type = vtReal;
        rval = d;
        size = 0;
    }

private:
    //No-fly zone
    Value(const Value&);
    Value& operator= (const Value&);
} ALGNL;
#pragma pack(pop)

}